/*
 * EPOC port for Level-9 interpreter by Simon Quinn
 * based on ncurses Unix interface port for Level-9 interpreter
 *
 * Unix Port Copyright (c) Jim Cameron 1998
 *
 * 9-Jun-1998 fixed bug in os_load_file
 *
 * A few notes on this port:
 *
 * I've modified level9.h to use the LITTLEENDIAN macro -- it assumes this if
 *	you're using DOS or Windows, otherwise it uses the byte sex macros. This
 *	shouldn't hurt a little-endian machine (Intel), but you must undefine
 *	LITTLEENDIAN if you've got a Motorola-type machine.
 *
 */

/*
 * Set this if you are compiling on a little-endian machine (ARM, Intel)
 */
#define LITTLEENDIAN 1

/*
 * Define this as 1 to get the Emacs-type key bindings
 *	Ctrl-A (go to beginning of line)
 *	Ctrl-B (back one character)
 *	Ctrl-D (delete character at cursor)
 *	Ctrl-E (go to end of line)
 *	Ctrl-F (forward one character)
 *	Ctrl-K (delete to end of line)
 */
#define EMACS_TYPE_KEYS 1

// EPOC
#include <e32base.h>
#include <e32cons.h>
#include <e32svr.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "level9.h"


#define CTRL_A '\x01'
#define CTRL_B '\x02'
#define CTRL_D '\x04'
#define CTRL_E '\x05'
#define CTRL_F '\x06'
#define CTRL_H '\x08'
#define CTRL_K '\x0B'


LOCAL_D CConsoleBase* console;

int   Line_width;
char *Line_buffer;
int   Line_ptr = 0;
int   Lines    = 0;
int   More_lines;

/*
 * This is used by the justification routine to hold the part of the input that
 *	wouldn't fit on the line
 */
char *Rump;

/*
 * See discussion with os_flush()
 */
int  Line_pos = 0;


/*
 * Justification kludge (see end of printline() )
 */
L9BOOL Suppress_newline = FALSE;


// Stores the name of the game including directory, used for creating
// a default save and restore file name
char gamefilename[256];


/*
 * History!
 */
//#define HISTSIZE 32
#define HISTSIZE 10
#define MAX_COMMAND_LENGTH 128

char  History_list [HISTSIZE] [MAX_COMMAND_LENGTH];
int   History_start;
int   History_end;

#define NEXT_HISTORY(h) { ++(h); if ((h) >= HISTSIZE) (h) = 0; }
#define PREV_HISTORY(h) { if ((h) == 0) (h) = HISTSIZE; --(h); }


extern "C" void os_sqprint(char *ibuff)
{
	console->Printf(_L("%s\n"),ibuff);
	console->Getch();
}

/*
 * Local function prototypes
 */
void printline (void);
static L9BOOL input_i (char *ibuff, int size, L9BOOL suppress_history);


/*
 * From porting.txt :
		os_printchar() prints a character to the output. The interface
		can either buffer this character or print it immediately, but
		if buffering is used then the characters must all be sent to the
		output when the interpreter calls os_flush(). A paragraph of
		text is output as one long stream of characters, without line
		breaks, so the interface must provide its own word wrapping, and
		any other features that are desired, such as justification or a
		[More] prompt. The carriage return character is always '\r',
		rather than '\n'.
 * Got that? Good
 */
void os_printchar(char c)
{
//console->Printf(_L("os_printchar %c %d\n"),c, Line_ptr);
//console->Getch();

  if (c == '\r')
  {

	/*
	 * Newline
	 */
	if (Line_ptr != 0)
	{
	  Suppress_newline = FALSE;
	}

	os_flush ();

	/*
	 * If the last line of the previous paragraph exactly filled the screen,
	 *	we now have a blank line and should not print a newline
	 */
	if (!Suppress_newline)
	{
	  console->Printf(_L("\n"));
	  ++Lines;
	}
	else
	{
	  Suppress_newline = FALSE;
	}

	Line_ptr = 0;

	Line_pos = 0;

  }
  else
  {
	Line_buffer [Line_ptr] = c;

	++Line_ptr;

	++Line_pos;
	if (Line_pos >= Line_width)
	{
	  Line_buffer [Line_ptr] = 0;
	  printline ();
	}

  }
  
  if (Lines == More_lines)
  {
	console->Printf(_L("[More]"));
	console->Getch();
	console->SetPos(0);
	console->Printf(_L("      "));
	console->SetPos(0);
	Lines = 0;
  }

}


/*
 * From porting.txt :
		os_input() reads a line of text from the user, usually to accept
		the next command to be sent to the game. The text input must be
		stored in ibuff with a terminating zero, and be no longer than
		size characters. Normally os_input() should return TRUE, but may
		return FALSE to cause the entire input so far to be discarded.
		The reason for doing so is discussed in the section at the end
		on allowing the interpreter to load a new game without exiting.
 *
 * The command-line history behaviour mirrors that of bash(1); if you edit a
 *	line and then move to another point in the history the current line will
 *	be saved in that history slot.
 *
 * Command-line editing works the way you'd expect (subject to $TERM's being
 *	set correctly and your termcap file's being kosher, of course). The input
 *	cursor takes up all the buffer space which doesn't actually contain
 *	characters; characters to the right of the cursor are stored at the very
 *	end of the input buffer. left_pos is the next position where a character
 *	is about to be stored, and right_pos points at the character under the
 *	cursor. This makes insertion much easier. (I got this trick from
 *	disassembling an old word-processor on the BBC micro).
 *
 * No, I don't have tab completion. Feel free to add it in ... 8-)
 */
L9BOOL os_input (char *ibuff, int size)
{

  return input_i (ibuff, size, FALSE);

}


/*
 * The 'i' stands for internal
 * The suppress_history parameter is used by the save/load filename input
 *	routines to allow command-line editing without history
 */
L9BOOL input_i (char *ibuff, int size, L9BOOL suppress_history)
{

  int left_pos = 0;
  int right_pos = size - 1;
  TKeyCode got;
//  int x, y;
  int oldx, oldy;

  int history_current = History_end;

  /*
   * A macro for fetching a history list entry
   */
# define FETCH_HISTORY								\
  console->SetPos(console->WhereX() - left_pos);	\
  console->ClearToEndOfLine();						\
  strcpy (ibuff, History_list [history_current]);	\
  left_pos = strlen (ibuff);						\
  console->Printf(_L("%s"), ibuff)

	ibuff [0] = '\0';

	oldx = console->WhereX();
	oldy = console->WhereY();

  /*
   * Player's got a chance to read all the text now
   */
  Lines = 0;

  /*
   * We are going to print a newline, but not through the normal channels.
   *  Make sure there's no dangling justification stuff
   */
  Line_pos = 0;

  if (size > MAX_COMMAND_LENGTH)
  {
	size = MAX_COMMAND_LENGTH;
  }

  /*
   * Clear the current entry in the history list
   */
  if (!suppress_history)
  {
	History_list [History_end] [0] = '\0';
  }

  while (1)
  {

	got = console->Getch();
	switch (got)
	{

	case '\n':
	case EKeyEnter:
		if ((strlen (ibuff) > 0) && !suppress_history)
		{
			strcpy (History_list [History_end], ibuff);
			NEXT_HISTORY (History_end);
			if (History_end == History_start)
			{
			  NEXT_HISTORY (History_start);
			}
		}
		console->Printf(_L("\n"));
		return TRUE;

	case EKeyBackspace:
	  if (left_pos != 0)
	  {
		--left_pos;
		/*
		 * CHANGEME This won't work when the cursor is at the beginning of a
		 *	line
		 */

		if(left_pos < (int)strlen(ibuff))
		{
			int a, savex, savey;
			for (a=left_pos ; a < (int)strlen(ibuff); a++)
				ibuff[a] = ibuff[a+1];
			console->SetPos(console->WhereX() - 1);
			savex = console->WhereX();
			savey = console->WhereY();
			console->Printf(_L("%s "), ibuff + left_pos);
			console->SetPos(savex, savey);
		}
		else
		{
			console->SetPos(console->WhereX() - 1);
			console->Write(_L(" "));
			console->SetPos(console->WhereX() - 1);
			ibuff[strlen(ibuff)] = '\0';
		}
	  }

	  break;

	case EKeyUpArrow:
	  if ((history_current == History_start) || suppress_history)
	  {
	  }
	  else
	  {
		strcpy (ibuff + left_pos, ibuff + right_pos);
		strcpy (History_list [history_current], ibuff);
		PREV_HISTORY (history_current);
		FETCH_HISTORY;
	  }
	  break;

	case EKeyDownArrow:
	  if ((history_current == History_end) || suppress_history)
	  {
	  }
	  else
	  {
		strcpy (ibuff + left_pos, ibuff + right_pos);
		strcpy (History_list [history_current], ibuff);
		NEXT_HISTORY (history_current);
		FETCH_HISTORY;
	  }
	  break;

	case EKeyLeftArrow:
	  if (left_pos > 0)
	  {
		console->SetPos(console->WhereX() - 1);
		--left_pos;
		--right_pos;
		ibuff [right_pos] = ibuff [left_pos];
	  }
	  break;

	case EKeyRightArrow:
		if(left_pos < (int)strlen(ibuff))
		{
			console->SetPos(console->WhereX() + 1);
			++left_pos;
		}
	  break;

	  /*
	   * Move to beginning of line
	   */
	case EKeyHome:
		left_pos = 0;
		console->SetPos(oldx, oldy);
	  break;

	  /*
	   * Move to end of line
	   */
	case EKeyEnd:
		left_pos = strlen(ibuff);
		console->SetPos(oldx + left_pos, oldy);
		break;

	//Everything else
	default:
		if(left_pos < (int)strlen(ibuff))
		{
			int a, savex, savey;
			ibuff[strlen(ibuff)+1] = 0;
			for (a=strlen(ibuff); a > left_pos; a--)
				ibuff[a] = ibuff[a-1];
			savex = console->WhereX();
			savey = console->WhereY();
			console->SetPos(oldx, oldy);
			ibuff [left_pos] = (char)got;
			console->Printf(_L("%s"), ibuff);
			console->SetPos(savex + 1, savey);
		}
		else
		{
			ibuff [left_pos] = (char) got;
			ibuff [left_pos + 1] = '\0';
			console->Printf(_L("%c"),got);
		}
		++left_pos;
		break;
	}
  }

# undef FETCH_HISTORY

  return FALSE;

}

char os_readchar(void)
{

	Lines = 0;

	return (char) console->Getch();

}

/*
 * From porting.txt :
	If the calls to os_printchar() are being buffered by the
	interface then the buffered text must be printed when os_flush()
	is called.
 */
void os_flush(void)
{
  Line_buffer [Line_ptr] = '\0';
  console->Printf(_L("%s"),Line_buffer);

  Line_ptr = 0;
}



/*
 * From porting.txt :
	os_save_file() should prompt the user in some way (with either
	text or a file requester) for a filename to save the area of
	memory of size Bytes pointed to by Ptr. TRUE or FALSE should be
	returned depending on whether or not the operation was successful.
 */
L9BOOL os_save_file(L9BYTE * Ptr, int Bytes)
{

  char fname [256], tempfilename[256];
  char *ext;

  FILE *outfile;

  // Create a default save file name from the loaded game file/directory
  tempfilename[0] = '\0';
  ext = strrchr(gamefilename,'.');
  if(ext != 0)
  {
	int templen;
	templen = strlen(gamefilename) - strlen(ext);
    strncpy(tempfilename, gamefilename, templen);
	tempfilename[templen] = '\0';
	strcat(tempfilename, ".sav"); 
  }

  os_flush();
  console->Printf(_L("\nSave file. "));
  console->Printf(_L("(Default %s): "),tempfilename);

  input_i (fname, 256, TRUE);

  if(strlen(fname) == 0)
    strcpy(fname, tempfilename);
	
  outfile = fopen (fname, "w");
  if (outfile == NULL)
  {
	return FALSE;
  }

  fwrite (Ptr, 1, Bytes, outfile);
  fclose (outfile);

  return TRUE;

}


/*
 * From porting.txt :
	os_load_file() should prompt the user for the name of a file to
	load. At most Max bytes should be loaded into the memory pointed
	to by Ptr, and the number of bytes read should be placed into the
	variable pointed to by Bytes.
 */
L9BOOL os_load_file(L9BYTE *Ptr,int *Bytes,int Max)
{
  char fname [256], tempfilename[256];
  char *ext;

  FILE *infile;

  // Create a default save file name from the loaded game file/directory
  tempfilename[0] = '\0';
  ext = strrchr(gamefilename,'.');
  if(ext != 0)
  {
	int templen;
	templen = strlen(gamefilename) - strlen(ext);
    strncpy(tempfilename, gamefilename, templen);
	tempfilename[templen] = '\0';
	strcat(tempfilename, ".sav"); 
  }

  os_flush();
  console->Printf(_L("\nLoad file. "));
  console->Printf(_L("(Default %s): "),tempfilename);

  input_i (fname, 256, TRUE);

  if(strlen(fname) == 0)
    strcpy(fname, tempfilename);

  infile = fopen (fname, "r");
  if (infile == NULL)
  {
	return FALSE;
  }

  *Bytes = fread (Ptr, 1, Max, infile);
  fclose (infile);
  return TRUE;

}


/*
 * From porting.txt :
	os_get_game_file() should prompt the user for a new game file, to
	be stored in NewName, which can take a maximum name of Size
	characters. This is used in the Adrian Mole games (and possibly
	others) which load in the next part of the game after the part
	currently being played has completed. These games were originally
	written for tape-based systems where the call was simply "load
	the next game from the tape".
 */
L9BOOL os_get_game_file(char *NewName,int Size)
{

  os_flush();
  console->Printf(_L("Load next game: "));
  return input_i (NewName, Size, TRUE);
  return TRUE;

}


/*
 * From porting.txt :
	os_set_filename() is for multi-part games originally written for
	disk-based systems, which used game filenames such as

		gamedat1.dat	gamedat2.dat

	etc. The routine should take the full filename in NewName (of
	maximum size Size) and modify it to reflect the number n, e.g.
	os_set_filename("gamedat1.dat",2) should leave "gamedat2.dat"
	in NewName.
 */
void os_set_filenumber(char *NewName,int /*Size*/,int n)
{

  char *leafname;
  int i;

  leafname = strrchr (NewName, FILE_DELIM);
  if (leafname == NULL)
  {
	leafname = NewName;
  }

  for (i = strlen (leafname) - 1; i >= 0; i--)
  {
	if (isdigit (leafname [i]))
	{
	  leafname [i] = '0' + n;
	  break;
	}
  }

}



/*
 * From porting.txt :
	You must provide your own main() entry point for the program.
	The simplest such main() is given in generic.c, which just calls
	LoadGame() and then sits in a loop calling RunGame(). These
	functions are discussed below.
 */
int lev_main (int argc, char *argv [])
{

//  int i;
  L9BOOL gotgame;

  /*
   * Check byte sex
   */
# if LITTLEENDIAN
  L9UINT32 test = 0x12345678;
  char *tcp = (char *) &test;
# else
  L9UINT32 test = 0x78563412;
  char *tcp = (char *) &test;
# endif
  if ((tcp [0] != 0x78) || (tcp [1] != 0x56) ||
	  (tcp [2] != 0x34) || (tcp [3] != 0x12))
  {
	console->Printf(_L("Compiled with the wrong byte sex! Check the LITTLEENDIAN\n"));
	exit (1);
  }

  if (argc != 2)
  {
	console->Printf(_L("Syntax: %s <gamefile>\n"),argv[0]);
	exit (1);
  }

	console->Printf(_L("Level 9 Interpreter v2.0 by Glen Summers\n"));
	console->Printf(_L("EPOC port by Simon Quinn\n\n"));

  /*
   * Get the terminal characteristics for line width and More prompt
   */
#ifdef __WINS__
	Line_width = 77;
	More_lines = 18;
#else
  Line_width  = console->ScreenSize().iWidth - 1;
  More_lines = console->ScreenSize().iHeight;
#endif
  More_lines--;

  Line_buffer = (char * ) malloc (2 * (Line_width + 1));
  if (Line_buffer == NULL)
  {
	console->Printf(_L("Couldn't allocate buffer for pretty-print\n\n"));
	exit (1);
  }
  Rump		  = Line_buffer + Line_width + 1;


  /*
   * Make sure the line buffer is terminated
   */
  Line_buffer [Line_width] = '\0';

  //Save the loaded game file name
  strcpy(gamefilename, argv[1]);

  gotgame = FALSE;
  if (!gotgame)
  {
	if (!LoadGame (argv [1]))
	{
	  console->Printf(_L("%s: couldn't open Level9 file %s\n"),argv [0], argv [1]);
	  exit (1);
	}
  }

  /*
   * Main game loop!
   */
  while (1)
  {
	if (!RunGame ())
	{
	  break;
	}
  }

  StopGame();
  FreeMemory();

  free (Line_buffer);

  console->Printf (_L("Thanks for playing Level9 ...\n\n"));
  console->Getch();

  return 0;

}




/*
 * Print the text in the line buffer up to the word in progress, justifying it
 *	if wanted
 */
void printline (void)
{

  char *spaceptr;
  char *hyphenptr;

  /*
   * ... otherwise ANY full-length line in a paragraph will cause the end-of-
   *  paragraph newline to be suppressed (oops!)
   */
  Suppress_newline = FALSE;

  ++Lines;

  spaceptr	= strrchr (Line_buffer, ' ');
  hyphenptr = strrchr (Line_buffer, '-');
  if ((spaceptr == NULL) && (hyphenptr == NULL))
  {

	/*
	 * Buffer contains no spaces - just print the whole line
	 */
	console->Printf(_L("%s\n"), Line_buffer);
	Line_ptr = 0;
	return;

  }

  if (spaceptr > hyphenptr)
  {
	*spaceptr = '\0';
	strcpy (Rump, spaceptr + 1);
  }
  else
  {
	strcpy (Rump, hyphenptr + 1);
	hyphenptr [1] = '\0';
  }

  console->Printf(_L("%s\n"), Line_buffer);

  strcpy (Line_buffer, Rump);
  Line_ptr = strlen (Line_buffer);

  Line_pos = Line_ptr;

  /*
   * If the last line of a paragraph exactly fills the screen, we will get an
   *  extra newline when attempting to flush the blank `next' line. Suppress
   *  this
   */
  if (Line_ptr == 0)
  {
	Suppress_newline = TRUE;
  }

}


/*
 * jim -- being Linux, we must provide stricmp
 */
int stricmp (const char *s1, const char *s2)
{
  int diff;

  while ((*s1 != '\0') || (*s2 != '\0'))
  {
	diff = toupper (*s1) - toupper (*s2);
	if (diff != 0)
	  return diff;
	++s1;
	++s2;
  }
  return 0;
}



void ConsoleMainL()
{

	//Process command line parameters
	char *argv[2] = {"","                                                             "};
	TInt numargs=2;
	RProcess me;
    TBuf<128> args=me.CommandLine();
	if(args.Length() == 0)
		numargs=1;
	args.ZeroTerminate();
	argv[1] = (char *) args.Ptr();

	TBuf<128> tempname = _L("Level9 ");
	tempname.Append(args);

	console = Console::NewL(tempname,
					TSize(KDefaultConsWidth, KDefaultConsHeight));
	CleanupStack::PushL(console);

	//Start main program
	lev_main(numargs, argv);

	CleanupStack::PopAndDestroy();

	CloseSTDLIB();
}

GLDEF_C TInt E32Main()
{
	__UHEAP_MARK;
	CTrapCleanup* cleanupStack = CTrapCleanup::New();
	TRAPD(error, ConsoleMainL());
	__ASSERT_ALWAYS(!error, User::Panic(_L("PEP"),error));
	delete cleanupStack;
	__UHEAP_MARKEND;
	return 0;
}

